import traceback
from collections import defaultdict

from sqlalchemy import null, or_, select
from sqlalchemy.orm import joinedload

from chellow.dloads import put_mem_val
from chellow.e.computer import (
    SiteSource,
    SupplySource,
    contract_func,
    displaced_era,
    forecast_date,
)
from chellow.models import Era, Session, Site, SiteEra, Source, Supply
from chellow.utils import c_months_u, ct_datetime_now, hh_max, hh_min


def totals_runner(mem_id, site_id, start_date):
    CATS = ("imp_net", "exp_net", "used", "displaced")
    QUANTS = (
        "kwh",
        "gbp",
        "p_per_kwh",
        "triad_gbp",
        "duos_gbp",
        "red_duos_gbp",
        "md_kw",
    )
    sess = None
    titles = ["Quantity"]
    titles.extend(CATS)
    log_list = []
    problem = ""

    jval = {"titles": titles, "rows": [], "status": "running", "log": log_list}
    try:
        with Session() as sess:
            site = Site.get_by_id(sess, site_id)

            def log(msg):
                return
                log_list.append(msg)
                put_mem_val(mem_id, jval)

            log("started thread")
            f_date = forecast_date()
            report_context = {}
            vals = dict((cat, defaultdict(int)) for cat in CATS)

            if start_date is None:
                now = ct_datetime_now()
                start_year = now.year - 1
                start_month = now.month
            else:
                start_year = start_date.year
                start_month = start_date.month

            for progress, (month_start, month_finish) in enumerate(
                c_months_u(start_year=start_year, start_month=start_month, months=12)
            ):
                jval["progress"] = progress
                put_mem_val(mem_id, jval)

                for era in sess.execute(
                    select(Era)
                    .join(SiteEra)
                    .join(Supply)
                    .join(Source)
                    .where(
                        SiteEra.site == site,
                        Era.start_date <= month_finish,
                        or_(Era.finish_date == null(), Era.finish_date >= month_start),
                        Era.imp_mpan_core != null(),
                        Source.code.in_(("net", "gen-net")),
                    )
                    .options(
                        joinedload(Era.imp_supplier_contract),
                        joinedload(Era.exp_supplier_contract),
                        joinedload(Era.mop_contract),
                        joinedload(Era.dc_contract),
                    )
                ).scalars():
                    chunk_start = hh_max(era.start_date, month_start)
                    chunk_finish = hh_min(era.finish_date, month_finish)

                    log("about to call first supplysource")
                    supply_ds = SupplySource(
                        sess,
                        chunk_start,
                        chunk_finish,
                        f_date,
                        era,
                        True,
                        report_context,
                    )

                    supplier_contract = era.imp_supplier_contract
                    import_vb_function = contract_func(
                        report_context, supplier_contract, "virtual_bill"
                    )
                    if import_vb_function is None:
                        raise Exception(
                            "Can't find the import_virtual_bill function in the "
                            "supplier contract. "
                        )
                    else:
                        import_vb_function(supply_ds)
                        v_bill = supply_ds.supplier_bill

                        if "problem" in v_bill and len(v_bill["problem"]) > 0:
                            raise Exception("Supplier Problem: " + v_bill["problem"])

                        try:
                            vals["imp_net"]["gbp"] += v_bill["net-gbp"]
                            if "triad-gbp" in v_bill:
                                vals["imp_net"]["triad_gbp"] += v_bill["triad-gbp"]
                            if "duos-red-gbp" in v_bill:
                                vals["imp_net"]["red_duos_gbp"] += v_bill[
                                    "duos-red-gbp"
                                ]
                            vals["imp_net"]["duos_gbp"] += sum(
                                v
                                for k, v in v_bill.items()
                                if k.endswith("-gbp") and k.startswith("duos-")
                            )
                        except KeyError:
                            raise Exception(
                                f"For the era {era.id} the virtual bill {v_bill} "
                                f"from the contract {supplier_contract} does not "
                                f"contain the net-gbp key."
                            )

                    dc_contract = era.dc_contract
                    supply_ds.contract_func(dc_contract, "virtual_bill")(supply_ds)
                    dc_bill = supply_ds.dc_bill
                    vals["imp_net"]["gbp"] += dc_bill["net-gbp"]
                    if "problem" in dc_bill and len(dc_bill["problem"]) > 0:
                        problem += "DC Problem: " + dc_bill["problem"]

                    mop_contract = era.mop_contract
                    mop_bill_function = supply_ds.contract_func(
                        mop_contract, "virtual_bill"
                    )
                    mop_bill_function(supply_ds)
                    mop_bill = supply_ds.mop_bill
                    vals["imp_net"]["gbp"] += mop_bill["net-gbp"]
                    if "problem" in mop_bill and len(mop_bill["problem"]) > 0:
                        problem += "MOP Problem: " + mop_bill["problem"]

                    exp_supplier_contract = era.exp_supplier_contract
                    if exp_supplier_contract is not None:
                        log("about to call second supply source")
                        exp_supply_ds = SupplySource(
                            sess,
                            chunk_start,
                            chunk_finish,
                            f_date,
                            era,
                            False,
                            report_context,
                        )
                        exp_vb_function = contract_func(
                            report_context, exp_supplier_contract, "virtual_bill"
                        )
                        if exp_vb_function is None:
                            problem += (
                                "Can't find the export_virtual_bill function in the "
                                "supplier contract. "
                            )
                        else:
                            exp_vb_function(exp_supply_ds)
                            v_bill = exp_supply_ds.supplier_bill

                            if "problem" in v_bill and len(v_bill["problem"]) > 0:
                                problem += (
                                    f"Export supplier Problem: {v_bill['problem']}"
                                )

                            try:
                                vals["exp_net"]["gbp"] += v_bill["net-gbp"]
                                if "triad-gbp" in v_bill:
                                    vals["exp_net"]["triad_gbp"] += v_bill["triad-gbp"]
                                if "duos-red-gbp" in v_bill:
                                    vals["exp_net"]["red_duos_gbp"] += v_bill[
                                        "duos-red-gbp"
                                    ]
                                vals["exp_net"]["duos_gbp"] += sum(
                                    v
                                    for k, v in v_bill.items()
                                    if k.endswith("-gbp") and k.startswith("duos-")
                                )
                            except KeyError:
                                problem += (
                                    f"For the era {era.id} the virtual bill "
                                    f"{v_bill} from the contract "
                                    f"{supplier_contract} does not contain the "
                                    f"net-gbp key."
                                )
                                problem += v_bill["problem"]

                disp_era = displaced_era(
                    sess, report_context, site, month_start, month_finish, f_date
                )
                log("about to call sitesource")
                site_ds = SiteSource(
                    sess,
                    site,
                    month_start,
                    month_finish,
                    f_date,
                    report_context,
                    disp_era,
                    exclude_virtual=False,
                )

                if disp_era is not None:
                    disp_supplier_contract = disp_era.imp_supplier_contract
                    disp_vb_function = contract_func(
                        report_context, disp_supplier_contract, "displaced_virtual_bill"
                    )
                    if disp_vb_function is None:
                        raise Exception(
                            f"The supplier contract {disp_supplier_contract} "
                            "doesn't have the displaced_virtual_bill() function."
                        )
                    disp_vb_function(site_ds)
                    disp_supplier_bill = site_ds.supplier_bill

                    try:
                        vals["displaced"]["gbp"] += disp_supplier_bill["net-gbp"]
                    except KeyError:
                        disp_supplier_bill["problem"] += (
                            f"For the supply {site_ds.mpan_core} the virtual bill "
                            f"{disp_supplier_bill} from the contract "
                            f"{disp_supplier_contract} does not contain the net-gbp "
                            f"key."
                        )
                    log(str(disp_supplier_bill))
                    if "triad-gbp" in disp_supplier_bill:
                        vals["displaced"]["triad_gbp"] += disp_supplier_bill[
                            "triad-gbp"
                        ]
                    if "duos-red-gbp" in disp_supplier_bill:
                        vals["displaced"]["red_duos_gbp"] += disp_supplier_bill[
                            "duos-red-gbp"
                        ]
                    vals["displaced"]["duos_gbp"] += sum(
                        v
                        for k, v in disp_supplier_bill.items()
                        if k.endswith("-gbp") and k.startswith("duos-")
                    )

                for hh in site.hh_data(
                    sess, month_start, month_finish, exclude_virtual=True
                ):
                    for cat in CATS:
                        vals_cat = vals[cat]
                        vals_cat["md_kw"] = max(vals_cat["md_kw"], hh[cat] * 2)
                        vals_cat["kwh"] += float(hh[cat])

            for quant in ("gbp", "triad_gbp", "duos_gbp", "red_duos_gbp"):
                vals["used"][quant] += vals["imp_net"][quant] + vals["displaced"][quant]

            for cat in CATS:
                vals_cat = vals[cat]
                if vals_cat["kwh"] > 0:
                    vals_cat["p_per_kwh"] = vals_cat["gbp"] / vals_cat["kwh"] * 100
                else:
                    vals_cat["p_per_kwh"] = 0

            FORMAT_LOOKUP = {
                "kwh": "{:,.0f}",
                "gbp": "{:,.2f}",
                "p_per_kwh": "{:,.2f}",
                "triad_gbp": "{:,.2f}",
                "duos_gbp": "{:,.2f}",
                "red_duos_gbp": "{:,.2f}",
                "md_kw": "{:,.2f}",
            }

            for quant in QUANTS:
                row = [quant]
                fmt = FORMAT_LOOKUP[quant]
                for cat in CATS:
                    row.append(fmt.format(vals[cat][quant]))
                jval["rows"].append(row)

    except BaseException:
        msg = traceback.format_exc()
        jval["problem"] = msg
    finally:
        jval["status"] = "finished"
        put_mem_val(mem_id, jval)



import sys
import threading
import traceback


from flask import g, redirect, request

import odio

from sqlalchemy import select
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.models import MeasurementRequirement, Session, Ssc, Tpr, User
from chellow.utils import req_bool


def write_spreadsheet(
    fl,
    compressed,
    ssc_rows,
    mr_rows,
    tpr_rows,
):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("SSCs", ssc_rows)
        f.append_table("MRs", mr_rows)
        f.append_table("TPRs", tpr_rows)


def content(
    user_id,
    compression,
):
    sess = rf = None
    ssc_rows = [
        [
            "ssc_code",
            "ssc_description",
            "is_import",
            "valid_from",
            "valid_to",
        ]
    ]
    mr_rows = [["ssc_code", "tpr_code"]]
    tpr_rows = [["tpr_code", "is_teleswitch", "is_gmt"]]
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)

            rf = open_file("sscs.ods", user, "wb")

            for ssc in sess.scalars(select(Ssc).order_by(Ssc.code)):
                ssc_rows.append(
                    (
                        ssc.code,
                        ssc.description,
                        ssc.is_import,
                        ssc.valid_from,
                        ssc.valid_to,
                    )
                )
            for mr in sess.scalars(
                select(MeasurementRequirement)
                .join(Ssc)
                .join(Tpr)
                .order_by(Ssc.code, Tpr.code)
                .options(
                    joinedload(MeasurementRequirement.ssc),
                    joinedload(MeasurementRequirement.tpr),
                )
            ):
                mr_rows.append((mr.ssc.code, mr.tpr.code))
            for tpr in sess.scalars(select(Tpr).order_by(Tpr.code)):
                tpr_rows.append((tpr.code, tpr.is_teleswitch, tpr.is_gmt))

            write_spreadsheet(
                rf,
                compression,
                ssc_rows,
                mr_rows,
                tpr_rows,
            )
    except BadRequest as e:
        msg = e.description + traceback.format_exc()
        sys.stderr.write(msg + "\n")
        ssc_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, ssc_rows, mr_rows, tpr_rows)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        ssc_rows.append(["Problem " + msg])
        if rf is None:
            msg = traceback.format_exc()
            ef = open_file("error.txt", None, "w")
            ef.write(msg + "\n")
            ef.close()
        else:
            write_spreadsheet(rf, compression, ssc_rows, mr_rows, tpr_rows)
    finally:
        if rf is not None:
            rf.close()


def do_get(sess):
    compression = req_bool("compression") if "compression" in request.values else True

    user = g.user

    args = (user.id, compression)
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



import csv
import sys
import threading
import traceback
from datetime import datetime as Datetime

from flask import g, redirect

from sqlalchemy import or_, text
from sqlalchemy.sql.expression import null

from chellow.dloads import open_file
from chellow.e.computer import SiteSource, contract_func, displaced_era, forecast_date
from chellow.models import Contract, Era, Session, Site, SiteEra, Source, Supply
from chellow.utils import c_months_u, hh_format, hh_range, req_int


def to_val(v):
    if isinstance(v, Datetime):
        return hh_format(v)
    elif isinstance(v, set):
        if len(v) == 1:
            return to_val(v.pop())
        else:
            return ""
    else:
        return str(v)


def content(contract_id, end_year, end_month, months, user):
    caches = {}
    f = None
    try:
        with Session() as sess:
            f = open_file("displaced.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            titles = [
                "Site Code",
                "Site Name",
                "Associated Site Ids",
                "From",
                "To",
                "Gen Types",
                "CHP kWh",
                "LM kWh",
                "Turbine kWh",
                "PV kWh",
            ]

            month_list = list(
                c_months_u(finish_year=end_year, finish_month=end_month, months=months)
            )
            start_date, finish_date = month_list[0][0], month_list[-1][-1]

            f_date = forecast_date()

            contract = Contract.get_supplier_by_id(sess, contract_id)
            sites = (
                sess.query(Site)
                .join(SiteEra)
                .join(Era)
                .join(Supply)
                .join(Source)
                .filter(
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                    Era.start_date <= finish_date,
                    or_(
                        Source.code.in_(("gen", "gen-grid")),
                        Era.exp_mpan_core != null(),
                    ),
                )
                .distinct()
            )
            bill_titles = contract_func(
                caches, contract, "displaced_virtual_bill_titles"
            )()

            for title in bill_titles:
                if title == "total-msp-kwh":
                    title = "total-displaced-msp-kwh"
                titles.append(title)
            writer.writerow(titles)

            for site in sites:
                for month_start, month_finish in month_list:
                    disp_era = displaced_era(
                        sess, caches, site, month_start, month_finish, f_date
                    )
                    if disp_era is None:
                        continue
                    supplier_contract = disp_era.imp_supplier_contract
                    if contract is not None and contract != supplier_contract:
                        continue

                    linked_sites = set()
                    generator_types = set()
                    for era in (
                        sess.query(Era)
                        .join(SiteEra)
                        .filter(
                            SiteEra.site == site,
                            Era.start_date <= month_finish,
                            or_(
                                Era.finish_date == null(),
                                Era.finish_date >= month_start,
                            ),
                        )
                    ):
                        for site_era in era.site_eras:
                            if site_era.site != site:
                                linked_sites.add(site_era.site.code)
                        supply = era.supply
                        if supply.generator_type is not None:
                            generator_types.add(supply.generator_type.code)

                    supply_ids = set()
                    for era in (
                        sess.query(Era)
                        .join(SiteEra)
                        .filter(
                            SiteEra.site == site,
                            SiteEra.is_physical,
                            Era.start_date <= month_finish,
                            or_(
                                Era.finish_date == null(),
                                Era.finish_date >= month_start,
                            ),
                        )
                    ):
                        supply_ids.add(era.supply.id)

                    vals = [
                        site.code,
                        site.name,
                        ", ".join(list(linked_sites)),
                        hh_format(month_start),
                        hh_format(month_finish),
                        ", ".join(list(generator_types)),
                    ]

                    total_gen_breakdown = {}

                    results = iter(
                        sess.execute(
                            text(
                                "select supply.id, hh_datum.value, "
                                "hh_datum.start_date, channel.imp_related, "
                                "source.code, generator_type.code as "
                                "gen_type_code from hh_datum, channel, source, "
                                "era, supply left outer join generator_type on "
                                "supply.generator_type_id = generator_type.id "
                                "where hh_datum.channel_id = channel.id and "
                                "channel.era_id = era.id and era.supply_id = "
                                "supply.id and supply.source_id = source.id and "
                                "channel.channel_type = 'ACTIVE' and not "
                                "(source.code = 'grid' and channel.imp_related "
                                "is true) and hh_datum.start_date >= "
                                ":month_start and hh_datum.start_date "
                                "<= :month_finish and "
                                "supply.id = any(:supply_ids) order "
                                "by hh_datum.start_date, supply.id"
                            ),
                            params={
                                "month_start": month_start,
                                "month_finish": month_finish,
                                "supply_ids": sorted(list(supply_ids)),
                            },
                        )
                    )

                    (
                        sup_id,
                        hh_val,
                        hh_start,
                        imp_related,
                        source_code,
                        gen_type_code,
                    ) = next(results, (None, None, None, None, None, None))

                    for hh_date in hh_range(caches, month_start, month_finish):
                        gen_breakdown = {}
                        exported = 0
                        while hh_start == hh_date:
                            if not imp_related and source_code in ("grid", "gen-grid"):
                                exported += hh_val
                            if (imp_related and source_code == "gen") or (
                                not imp_related and source_code == "gen-grid"
                            ):
                                gen_breakdown[gen_type_code] = (
                                    gen_breakdown.setdefault(gen_type_code, 0) + hh_val
                                )

                            if (not imp_related and source_code == "gen") or (
                                imp_related and source_code == "gen-grid"
                            ):
                                gen_breakdown[gen_type_code] = (
                                    gen_breakdown.setdefault(gen_type_code, 0) - hh_val
                                )

                            (
                                sup_id,
                                hh_val,
                                hh_start,
                                imp_related,
                                source_code,
                                gen_type_code,
                            ) = next(results, (None, None, None, None, None, None))

                        displaced = sum(gen_breakdown.values()) - exported
                        added_so_far = 0
                        for key in sorted(gen_breakdown.keys()):
                            kwh = gen_breakdown[key]
                            if displaced < 0:
                                total_gen_breakdown[key] = (
                                    total_gen_breakdown.get(key, 0) + kwh
                                )
                            else:
                                if kwh + added_so_far > displaced:
                                    total_gen_breakdown[key] = (
                                        total_gen_breakdown.get(key, 0)
                                        + displaced
                                        - added_so_far
                                    )
                                    break
                                else:
                                    total_gen_breakdown[key] = (
                                        total_gen_breakdown.get(key, 0) + kwh
                                    )
                                    added_so_far += kwh

                    for title in ["chp", "lm", "turb", "pv"]:
                        vals.append(str(total_gen_breakdown.get(title, "")))

                    site_ds = SiteSource(
                        sess,
                        site,
                        month_start,
                        month_finish,
                        f_date,
                        caches,
                        disp_era,
                    )
                    disp_func = contract_func(
                        caches, supplier_contract, "displaced_virtual_bill"
                    )
                    disp_func(site_ds)
                    bill = site_ds.supplier_bill
                    for title in bill_titles:
                        if title in bill:
                            vals.append(to_val(bill[title]))
                            del bill[title]
                        else:
                            vals.append("")

                    for k in sorted(bill.keys()):
                        vals.append(k)
                        vals.append(str(bill[k]))
                    writer.writerow(vals)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        if f is not None:
            f.write(msg)
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    end_year = req_int("finish_year")
    end_month = req_int("finish_month")
    months = req_int("months")
    contract_id = req_int("supplier_contract_id")
    args = contract_id, end_year, end_month, months, g.user
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



import csv
import threading
import traceback
from io import StringIO

from flask import g, redirect, request

from sqlalchemy import or_, select
from sqlalchemy.sql.expression import null

from chellow.dloads import open_file
from chellow.e.computer import SupplySource, forecast_date
from chellow.models import Channel, Era, HhDatum, RSession, Supply, User
from chellow.utils import (
    csv_make_val,
    ct_datetime,
    hh_range,
    parse_mpan_core,
    req_bool,
    req_date,
    req_int,
    req_str,
    to_ct,
    to_utc,
)


def _write_row(writer, titles, vals, total):
    vals["total"] = total
    writer.writerow(csv_make_val(vals[t]) for t in titles)


def content(
    start_date,
    finish_date,
    imp_related,
    channel_type,
    is_zipped,
    supply_id,
    mpan_cores,
    user_id,
):
    finish_date_ct = to_ct(finish_date)
    zf = sess = tf = None
    base_name = ["supplies_hh_data", finish_date_ct.strftime("%Y%m%d%H%M")]
    cache = {}
    try:
        with RSession() as sess:
            supplies = (
                select(Supply)
                .join(Era)
                .where(
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                    Era.start_date <= finish_date,
                )
                .order_by(Supply.id)
                .distinct()
            )
            if supply_id is not None:
                supply = Supply.get_by_id(sess, supply_id)
                supplies = supplies.where(Supply.id == supply.id)
                first_era = sess.scalars(
                    select(Era)
                    .where(
                        Era.supply == supply,
                        or_(Era.finish_date == null(), Era.finish_date >= start_date),
                        Era.start_date <= finish_date,
                    )
                    .order_by(Era.start_date)
                ).first()
                if first_era.imp_mpan_core is None:
                    name_core = first_era.exp_mpan_core
                else:
                    name_core = first_era.imp_mpan_core
                base_name.append("supply_" + name_core.replace(" ", "_"))

            if mpan_cores is not None:
                supplies = supplies.where(
                    or_(
                        Era.imp_mpan_core.in_(mpan_cores),
                        Era.exp_mpan_core.in_(mpan_cores),
                    )
                )
                base_name.append("filter")

            cf = StringIO()
            writer = csv.writer(cf, lineterminator="\n")
            headers = [
                "import_mpan_core",
                "export_mpan_core",
                "is_hh",
                "is_import_related",
                "channel_type",
                "hh_start_clock_time",
                "total",
            ]
            slot_names = [str(n) for n in range(1, 51)]
            titles = headers + slot_names
            writer.writerow(titles)
            titles_csv = cf.getvalue()
            cf.close()
            fdate = forecast_date()
            user = User.get_by_id(sess, user_id)

            if is_zipped:
                zf = open_file(
                    "_".join(base_name) + ".zip", user, mode="w", is_zip=True
                )
            else:
                tf = open_file("_".join(base_name) + ".csv", user, mode="w", newline="")
                tf.write(titles_csv)

            for supply in sess.execute(supplies).scalars():
                cf = StringIO()
                writer = csv.writer(cf, lineterminator="\n")
                imp_mpan_core = exp_mpan_core = is_hh = None
                data = []
                for era in sess.scalars(
                    select(Era)
                    .where(
                        Era.supply == supply,
                        or_(Era.finish_date == null(), Era.finish_date >= start_date),
                        Era.start_date <= finish_date,
                    )
                    .order_by(Era.start_date)
                ):
                    if era.imp_mpan_core is not None:
                        imp_mpan_core = era.imp_mpan_core
                    if era.exp_mpan_core is not None:
                        exp_mpan_core = era.exp_mpan_core

                    if era.pc.code == "00" or len(era.channels) > 0:
                        is_hh = True
                        for datum in sess.scalars(
                            select(HhDatum)
                            .join(Channel)
                            .join(Era)
                            .where(
                                Era.supply == supply,
                                HhDatum.start_date >= start_date,
                                HhDatum.start_date <= finish_date,
                                Channel.imp_related == imp_related,
                                Channel.channel_type == channel_type,
                            )
                            .order_by(HhDatum.start_date)
                        ):
                            data.append((datum.start_date, datum.value))
                    else:
                        is_hh = False
                        ds = SupplySource(
                            sess,
                            start_date,
                            finish_date,
                            fdate,
                            era,
                            imp_related,
                            cache,
                        )
                        KEY_LOOKUP = {
                            "ACTIVE": "msp-kwh",
                            "REACTIVE_IMP": "imp-msp-kvarh",
                            "REACTIVE_EXP": "exp-msp-kvarh",
                        }
                        for hh in ds.hh_data:
                            data.append(
                                (hh["start-date"], hh[KEY_LOOKUP[channel_type]])
                            )

                vals = total = slot_count = None
                hh_data = iter(data)
                datum = next(hh_data, None)
                for current_date in hh_range(cache, start_date, finish_date):
                    dt_ct = to_ct(current_date)
                    if dt_ct.hour == 0 and dt_ct.minute == 0:
                        if vals is not None:
                            _write_row(writer, titles, vals, total)

                        vals = {
                            "import_mpan_core": imp_mpan_core,
                            "export_mpan_core": exp_mpan_core,
                            "is_hh": is_hh,
                            "is_import_related": imp_related,
                            "channel_type": channel_type,
                            "hh_start_clock_time": dt_ct.strftime("%Y-%m-%d"),
                        }
                        for slot_name in slot_names:
                            vals[slot_name] = None
                        total = slot_count = 0

                    slot_count += 1
                    if datum is not None and datum[0] == current_date:
                        val = datum[1]
                        vals[str(slot_count)] = val
                        total += val
                        datum = next(hh_data, None)

                if vals is not None:
                    _write_row(writer, titles, vals, total)

                if is_zipped:
                    fname = f"{imp_mpan_core}_{exp_mpan_core}_{supply.id}.csv".replace(
                        " ", "_"
                    )
                    zf.writestr(fname, titles_csv + cf.getvalue())
                else:
                    tf.write(cf.getvalue())
                cf.close()

                # Avoid long-running transaction
                sess.rollback()

            if is_zipped:
                zf.close()
            else:
                tf.close()

    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if is_zipped:
            zf.writestr("error.txt", msg)
            zf.close()
        else:
            tf.write(msg)


def do_get(sess):
    return handle_request()


def do_post(sess):
    if "mpan_cores" in request.values:
        mpan_cores_str = req_str("mpan_cores")
        mpan_cores = mpan_cores_str.splitlines()
        if len(mpan_cores) == 0:
            mpan_cores = None
        else:
            for i in range(len(mpan_cores)):
                mpan_cores[i] = parse_mpan_core(mpan_cores[i])
    return handle_request(mpan_cores)


def handle_request(mpan_cores=None):
    start_date = req_date("start", resolution="day")

    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    finish_day = req_int("finish_day")

    finish_date = to_utc(ct_datetime(finish_year, finish_month, finish_day, 23, 30))

    imp_related = req_bool("imp_related")
    channel_type = req_str("channel_type")
    is_zipped = req_bool("is_zipped")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    user = g.user
    args = (
        start_date,
        finish_date,
        imp_related,
        channel_type,
        is_zipped,
        supply_id,
        mpan_cores,
        user.id,
    )
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



from decimal import Decimal
from io import StringIO

from chellow.e.hh_importer import (
    HhDataImportProcess,
    https_handler,
    solaredge_energy_details_parser,
)
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_https_handler(mocker, sess):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", valid_from, None, None
    )
    participant.insert_party(sess, market_role_X, "Fusion Ltc", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, valid_from, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, valid_from, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", valid_from, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, valid_from, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        valid_from,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, valid_from, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    era.insert_channel(sess, True, "ACTIVE")

    sess.commit()

    mock_requests = mocker.patch("chellow.e.hh_importer.requests")
    mock_s = mocker.Mock()
    mock_requests.Session.return_value = mock_s
    mock_response = mocker.Mock()
    mock_s.get.return_value = mock_response
    mock_response.json.return_value = {
        "Data": [{"Flags": 0, "Time": 636188256000000000, "Value": 21}]
    }

    log = []

    def log_f(msg):
        log.append(msg)

    properties = {
        "enabled": True,
        "protocol": "https",
        "download_days": 8,
        "parser": "meniscus",
        "url_template": "https://example.com/?from="
        "{{chunk_start.strftime('%d/%m/%Y')}}&to="
        "{{chunk_finish.strftime('%d/%m/%Y')}}",
        "url_values": {"22 7907 4116 080": {"api_key": "768234ht"}},
    }

    now = utc_datetime(2020, 12, 22)

    https_handler(sess, log_f, properties, dc_contract, now=now)
    expected_log = [
        "Window start: 2020-12-14 00:00",
        "Window finish: 2020-12-21 23:30",
        "Looking at MPAN core 22 7867 6232 781.",
        "Retrieving data from https://example.com/?from=14/12/2020&to=21/12/2020.",
        "Finished loading.",
    ]
    assert log == expected_log
    mock_s.get.assert_called_once()


def test_HhDataImportProcess(sess):
    valid_from = to_utc(ct_datetime(1996, 1, 1))

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AB Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "FDC Ltd.", valid_from, None, None)
    dc_contract = Contract.insert_dc(
        sess,
        "Fusion DC 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    sess.commit()

    process_id = 0
    content = "#F2"
    istream = StringIO(content)
    file_name = "afile.df2"
    file_size = len(content)
    importer = HhDataImportProcess(
        dc_contract.id, process_id, istream, file_name, file_size
    )
    importer.run()

    assert importer.messages == []


def test_solaredge_energy_details_parser():
    def log_f():
        pass

    mpan_core = "22 4781 4783 280"
    res = {
        "energyDetails": {
            "meters": [
                {
                    "type": "Production",
                    "values": [{"date": "2023-10-01 00:00:00", "value": "3772"}],
                }
            ]
        }
    }
    properties = {"generated_mpan_cores": [mpan_core]}
    actual = solaredge_energy_details_parser(log_f, res, properties, mpan_core)
    expected = [
        {
            "mpan_core": "22 4781 4783 280",
            "start_date": utc_datetime(2023, 9, 30, 23, 0),
            "channel_type": "ACTIVE",
            "value": Decimal("3.772"),
            "status": "A",
        }
    ]
    print(actual)
    assert actual == expected



import csv
import sys
import threading
import traceback

from flask import g, redirect

from chellow.dloads import open_file
from chellow.models import Session, Site, User
from chellow.utils import c_months_u, csv_make_val, req_int, req_str, to_ct


def _write_row(writer, total, values, titles):
    values["total"] = total
    writer.writerow([csv_make_val(values[t]) for t in titles])


def content(start_date, finish_date, site_id, typ, user_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file(
                f'site_hh_data_{to_ct(start_date).strftime("%Y%m%d%H%M")}.csv',
                user,
                mode="w",
                newline="",
            )
            writer = csv.writer(f, lineterminator="\n")
            titles = ["site_code", "type", "hh_start_clock_time", "total"]
            hr_titles = tuple(map(str, range(1, 51)))
            titles.extend(hr_titles)
            writer.writerow(titles)
            site = Site.get_by_id(sess, site_id)

            vals = total = hh_num = None
            for hh in site.hh_data(sess, start_date, finish_date):
                hh_start_ct = to_ct(hh["start_date"])
                if (hh_start_ct.hour, hh_start_ct.minute) == (0, 0):
                    if vals is not None:
                        _write_row(writer, total, vals, titles)
                    vals = {
                        "site_code": site.code,
                        "type": typ,
                        "hh_start_clock_time": hh_start_ct.strftime("%Y-%m-%d"),
                    }
                    for t in hr_titles:
                        vals[t] = None
                    total = 0
                    hh_num = 1
                val = hh[typ]
                vals[str(hh_num)] = val
                hh_num += 1
                total += val
            if vals is not None:
                _write_row(writer, total, vals, titles)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    months = req_int("months")
    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")

    month_list = list(
        c_months_u(finish_year=finish_year, finish_month=finish_month, months=months)
    )
    start_date, finish_date = month_list[0][0], month_list[-1][-1]

    typ = req_str("type")
    site_id = req_int("site_id")
    args = start_date, finish_date, site_id, typ, g.user.id
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



from chellow.e.bill_parsers.settlement_dc_stark_xlsx import Parser


def test(sess):
    with open(
        "test/e/bill_parsers/test_bill_parser_settlement_dc_stark_xlsx/"
        "bills.settlement.dc.stark.xlsx",
        mode="rb",
    ) as f:
        parser = Parser(f)
        parser.make_raw_bills()



import csv
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from sqlalchemy import or_, select
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from zish import dumps, loads

from chellow.models import Contract, RateScript
from chellow.utils import HH, ct_datetime, hh_format, hh_range, to_ct, to_utc


RUNS = ["DF", "RF", "R3", "R2", "R1", "SF", "II"]


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source, run="DF"):
    gsp_group_code = data_source.gsp_group_code
    use_runs = RUNS[RUNS.index(run) :]

    try:
        cache = data_source.caches["tlms"]
    except KeyError:
        cache = data_source.caches["tlms"] = {}

    for h in data_source.hh_data:
        try:
            h["tlm"] = tlm = cache[h["start-date"]][gsp_group_code][run]
        except KeyError:
            h_start = h["start-date"]
            vals = data_source.non_core_rate("tlms", h_start, exact=True)
            if vals is None:
                vals = data_source.non_core_rate("tlms", h["hist-start"])

            rates = vals["tlms"]

            key = key_format(h_start)
            try:
                rate = rates[key]
            except KeyError:
                rate = sorted(rates.items())[-1][1]

            try:
                gsp_rate = rate[gsp_group_code]
            except KeyError:
                gsp_rate = sorted(rate.items())[-1][1]

            tlm = None
            for use_run in use_runs:
                try:
                    h["tlm"] = tlm = float(gsp_rate[use_run]["off_taking"])
                    break
                except KeyError:
                    pass

            if tlm is None:
                h["tlm"] = tlm = float(sorted(gsp_rate.items())[-1][1]["off_taking"])

            try:
                rates_cache = cache[h["start-date"]]
            except KeyError:
                rates_cache = cache[h["start-date"]] = {}

            try:
                gsp_cache = rates_cache[gsp_group_code]
            except KeyError:
                gsp_cache = rates_cache[gsp_group_code] = {}

            gsp_cache[run] = tlm

        h["nbp-kwh"] = h["gsp-kwh"] * tlm


def _find_complete_date(caches, sess, contract, cache):
    for rs in sess.scalars(
        select(RateScript)
        .where(RateScript.contract == contract, RateScript.finish_date != null())
        .order_by(RateScript.start_date.desc())
    ):
        rates = rs.make_script()
        rates["id"] = rs.id
        cache["rate_scripts"].append(rates)
        timestamps = cache["timestamps"]
        tlms = rates["tlms"]
        complete = True
        for dt in hh_range(caches, rs.start_date, rs.finish_date):
            timestamps[dt] = rates
            key = key_format(dt)
            if key in tlms:
                gps = tlms[key]
                for group in GSP_GROUP_LOOKUP.values():
                    if not (group in gps and "DF" in gps[group]):
                        complete = False
            else:
                complete = False

        if complete:
            return rs.finish_date


def elexon_import(sess, log, set_progress, s, scripting_key):
    cache = {"rate_scripts": [], "timestamps": {}}
    caches = {}
    log("Starting to check TLMs.")
    contract_name = "tlms"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1997, 1, 1)),
            to_utc(ct_datetime(1997, 1, 31, 23, 30)),
            {"tlms": {}},
        )
        sess.commit()
    contract_props = contract.make_properties()
    if contract_props.get("enabled", False):

        params = {"key": scripting_key}
        url_str = "https://downloads.elexonportal.co.uk/file/download/TLM_FILE"
        complete_date = _find_complete_date(caches, sess, contract, cache)
        log(f"Found complete up to {complete_date}")

        sess.rollback()  # Avoid long-running transaction
        r = s.get(url_str, params=params, timeout=120)
        parser = csv.reader(
            (x.decode() for x in r.iter_lines()), delimiter=",", quotechar='"'
        )
        log(f"Opened {url_str}?key={scripting_key}.")

        next(parser, None)
        for values in parser:
            if len(values) == 0:
                continue

            set_progress(values[0])

            if values[3] == "":
                for zone in GSP_GROUP_LOOKUP.keys():
                    values[3] = zone
                    _process_line(
                        cache, sess, contract, log, values, complete_date, caches
                    )
            else:
                _process_line(cache, sess, contract, log, values, complete_date, caches)

        for rscript in cache["rate_scripts"]:
            rs = sess.scalars(
                select(RateScript).where(RateScript.id == rscript["id"])
            ).one()
            rs.script = dumps({"tlms": rscript["tlms"]})
            sess.commit()

    else:
        log(
            "The importer is disabled. Set 'enabled' to 'true' in the "
            "properties to enable it."
        )
    log("Finished checking TLMs.")


GSP_GROUP_LOOKUP = {
    "1": "_A",
    "2": "_B",
    "3": "_C",
    "4": "_D",
    "5": "_E",
    "6": "_F",
    "7": "_G",
    "8": "_H",
    "9": "_J",
    "10": "_K",
    "11": "_L",
    "12": "_M",
    "13": "_N",
    "14": "_P",
}


def _process_line(cache, sess, contract, log_func, values, complete_date, caches):
    hh_date_ct = to_ct(Datetime.strptime(values[0], "%d/%m/%Y"))
    hh_date = to_utc(hh_date_ct)
    hh_date += relativedelta(minutes=30 * (int(values[2]) - 1))
    if complete_date is not None and hh_date <= complete_date:
        return
    run = values[1]
    gsp_group_code = GSP_GROUP_LOOKUP[values[3]]
    off_taking_str = values[4]

    try:
        off_taking = Decimal(off_taking_str)
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing 'off-taking' field '{off_taking_str}' in the row "
            f"{values}. {e}"
        )

    delivering = Decimal(values[5])

    key = key_format(hh_date)
    try:
        cache["timestamps"][hh_date]["tlms"][key][gsp_group_code][run]
    except KeyError:
        try:
            rate = cache["timestamps"][hh_date]
        except KeyError:

            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= hh_date,
                    or_(
                        RateScript.finish_date == null(),
                        RateScript.finish_date >= hh_date,
                    ),
                )
                .first()
            )
            assert rs is None
            while rs is None:
                log_func(f"There's no rate script at {hh_format(hh_date)}.")
                latest_rs = (
                    sess.query(RateScript)
                    .filter(RateScript.contract == contract)
                    .order_by(RateScript.start_date.desc())
                    .first()
                )
                contract.update_rate_script(
                    sess,
                    latest_rs,
                    latest_rs.start_date,
                    latest_rs.start_date + relativedelta(months=2) - HH,
                    loads(latest_rs.script),
                )
                new_rs_start = latest_rs.start_date + relativedelta(months=1)
                new_rs = contract.insert_rate_script(sess, new_rs_start, {})
                rt = {"tlms": {}, "id": new_rs.id}
                cache["rate_scripts"].append(rt)
                for h in hh_range(caches, new_rs.start_date, new_rs.finish_date):
                    cache["timestamps"][h] = rt
                sess.commit()
                log_func(f"Added a rate script starting at {hh_format(new_rs_start)}.")

                rs = (
                    sess.query(RateScript)
                    .filter(
                        RateScript.contract == contract,
                        RateScript.start_date <= hh_date,
                        or_(
                            RateScript.finish_date == null(),
                            RateScript.finish_date >= hh_date,
                        ),
                    )
                    .first()
                )

            rate = cache["timestamps"][hh_date]

        try:
            existing = rate["tlms"][key]
        except KeyError:
            existing = rate["tlms"][key] = {}

        try:
            group = existing[gsp_group_code]
        except KeyError:
            group = existing[gsp_group_code] = {}

        group[run] = {"off_taking": off_taking, "delivering": delivering}

        log_func(
            f"Found rate at {hh_format(hh_date)} for GSP Group {gsp_group_code} "
            f"and run {run}."
        )



from io import BytesIO
from zipfile import ZipFile

from requests.auth import _basic_auth_str

from utils import match

from chellow.models import Contract, MarketRole, Participant, Site, User, UserRole
from chellow.reports.report_183 import _process_site, none_content
from chellow.utils import ct_datetime, to_utc


def test_do_post(mocker, client, sess):
    user = User.get_by_email_address(sess, "admin@example.com")
    MockThread = mocker.patch("chellow.reports.report_183.threading.Thread")

    data = {
        "start_year": "2022",
        "start_month": "07",
        "start_day": "01",
        "finish_year": "2022",
        "finish_month": "07",
        "finish_day": "31",
        "type": "used",
        "site_codes": "",
    }
    response = client.post("/reports/183", data=data)

    match(response, 303)

    expected_args = (
        None,
        "used",
        to_utc(ct_datetime(2022, 7, 1)),
        to_utc(ct_datetime(2022, 7, 31, 23, 30)),
        user.id,
        "sites_hh_data_202207312330_filter.zip",
    )

    MockThread.assert_called_with(target=none_content, args=expected_args)


def test_do_post_site_codes(mocker, client, sess):
    user = User.get_by_email_address(sess, "admin@example.com")
    MockThread = mocker.patch("chellow.reports.report_183.threading.Thread")

    data = {
        "start_year": "2022",
        "start_month": "07",
        "start_day": "01",
        "finish_year": "2022",
        "finish_month": "07",
        "finish_day": "31",
        "type": "used",
        "site_codes": "8ghh3",
    }
    response = client.post("/reports/183", data=data)

    match(response, 303)

    expected_args = (
        ["8ghh3"],
        "used",
        to_utc(ct_datetime(2022, 7, 1)),
        to_utc(ct_datetime(2022, 7, 31, 23, 30)),
        user.id,
        "sites_hh_data_202207312330_filter.zip",
    )

    MockThread.assert_called_with(target=none_content, args=expected_args)


def test_do_post_viewer(mocker, raw_client, sess):
    vf = to_utc(ct_datetime(2022, 7, 1))
    user_role = UserRole.insert(sess, "viewer")
    user = User.insert(sess, "admin@example.com", "admin", user_role, None)
    participant = Participant.insert(sess, "CALB", "Calb")
    market_role = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role, "neut", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    sess.commit()

    MockThread = mocker.patch("chellow.reports.report_183.threading.Thread")

    data = {
        "start_year": "2022",
        "start_month": "07",
        "start_day": "01",
        "finish_year": "2022",
        "finish_month": "07",
        "finish_day": "31",
        "type": "used",
        "site_codes": "",
    }
    headers = {"Authorization": _basic_auth_str("admin@example.com", "admin")}
    response = raw_client.post("/reports/183", data=data, headers=headers)

    match(response, 303)

    expected_args = (
        None,
        "used",
        to_utc(ct_datetime(2022, 7, 1)),
        to_utc(ct_datetime(2022, 7, 31, 23, 30)),
        user.id,
        "sites_hh_data_202207312330_filter.zip",
    )

    MockThread.assert_called_with(target=none_content, args=expected_args)


def test_process_site(sess):
    f = BytesIO()
    zf = ZipFile(f, mode="w")
    site = Site.insert(sess, "CI017", "Water Works")
    start_date = to_utc(ct_datetime(2022, 7, 1))
    finish_date = to_utc(ct_datetime(2022, 7, 31, 23, 30))
    typ = "used"
    _process_site(sess, zf, site, start_date, finish_date, typ)
    assert zf.namelist() == ["CI017_202207312330.csv"]


def test_none_content(mocker, sess):
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    sess.commit()

    f = BytesIO()
    mocker.patch("chellow.reports.report_183.open_file", return_value=f)

    site_codes = []
    typ = "used"
    start_date = to_utc(ct_datetime(2022, 7, 1))
    finish_date = to_utc(ct_datetime(2022, 7, 31, 23, 30))
    file_name = "output"
    none_content(site_codes, typ, start_date, finish_date, user_id, file_name)



import datetime
from collections import defaultdict
from decimal import Decimal
from io import StringIO

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.utils import parse_mpan_core, to_ct, to_utc


def parse_date(date_str):
    return to_utc(to_ct(parse_date_naive(date_str)))


def parse_date_naive(date_str):
    return datetime.datetime.strptime(date_str, "%Y%m%d")


def _chop_record(record, **kwargs):
    parts = {}
    idx = 0
    for name, length in kwargs.items():
        parts[name] = record[idx : idx + length]
        idx += length
    return parts


DATE_LENGTH = 8


def _handle_0000(headers, pre_record, record):
    parts = _chop_record(record, unknown_1=9, issue_date=DATE_LENGTH)
    headers["issue_date"] = parse_date(parts["issue_date"])


def _handle_0050(headers, pre_record, record):
    parts = _chop_record(
        record,
        issue_date=DATE_LENGTH,
        unknown_1=12,
        unknown_2=12,
        unknown_3=12,
        unknown_4=12,
        late_payment=12,
    )
    headers["late_payment"] = Decimal(parts["late_payment"]) / Decimal(100)


def _handle_0051(headers, pre_record, record):
    pass


def _handle_0100(headers, pre_record, record):
    issue_date = headers["issue_date"]
    late_payment = headers.get("late_payment")
    headers.clear()
    headers["issue_date"] = issue_date
    if late_payment is not None:
        headers["late_payment"] = late_payment
    headers["account"] = pre_record[33:41]
    headers["reference"] = pre_record[41:46]
    headers["kwh"] = Decimal("0")
    headers["breakdown"] = defaultdict(int, {"vat": {}})
    headers["reads"] = []


def _handle_0101(headers, pre_record, record):
    parts = _chop_record(record, start_date=DATE_LENGTH, finish_date=DATE_LENGTH)
    headers["start_date"] = parse_date(parts["start_date"])
    headers["finish_date"] = to_utc(
        to_ct(
            parse_date_naive(parts["finish_date"]) + relativedelta(hours=23, minutes=30)
        )
    )


CHARGE_UNITS_LOOKUP = {
    "STDG": "days",
    "UNIT": "kwh",
    "AVAL": "kva",
    "EXAVAL": "kva",
    "MD": "kw",
    "LOADU": "kw",
    "SAG": "days",
    "TNUOS": "days",
    "REAP": "kvarh",
    "DCDA": "days",
    "MOP1": "days",
    "COMM": "days",
}

ELEMENT_LOOKUP = {
    "1ANNUAL": "duos-red",
    "2ANNUAL": "duos-amber",
    "3ANNUAL": "duos-green",
    "7ANNUAL": "duos-fixed",
    "9ANNUAL": "duos-reactive",
    "10ANNUAL": "standing",
    "20RS0108": "single",
    "30ANNUAL": "mop",
    "50ANNUAL": "dc",
    "9WANNUAL": "site-fee",
    "20RS0123": "day",
    "30RS0123": "night",
    "90ANNUAL": "duos-fixed",
    "9QANNUAL": "duos-availability",
    "9UANNUAL": "tnuos",
    "40ANNUAL": "maximum-demand",
    "20ANNUAL": "triad",
    "70ANNUAL": "elexon",
    "10RS0050": "duos-red",
    "20RS0050": "duos-amber",
    "30RS0050": "duos-red",
    "9CANNUAL": "duos-reactive",
    "40RS0050": "duos-super-red",
}


def _handle_0460(headers, pre_record, record):
    parts = _chop_record(
        record,
        unknown_1=12,
        unknown_2=12,
        code=8,
        quantity=12,
        units=22,
        rate=16,
        unknown_date=DATE_LENGTH,
        gbp=12,
        charge_description=35,
        unknown_3=51,
        days=2,
    )
    units = CHARGE_UNITS_LOOKUP[parts["units"].strip()]
    gbp = Decimal(parts["gbp"]) / 100
    quantity = Decimal(parts["quantity"])
    rate = Decimal(parts["rate"])
    element_name = ELEMENT_LOOKUP[parts["code"].strip()]
    breakdown = headers["breakdown"]
    breakdown[f"{element_name}-{units}"] += quantity
    rate_name = f"{element_name}-rate"
    if rate_name in breakdown:
        rates = breakdown[rate_name]
    else:
        rates = breakdown[rate_name] = set()

    rates.add(rate)
    breakdown[f"{element_name}-gbp"] += gbp
    if element_name in ("duos-availability", "duos-excess-availability"):
        breakdown[f"{element_name}-days"] += Decimal(parts["days"])


CONSUMPTION_UNITS_LOOKUP = {"KWH": "kWh", "KVA": "kVA", "KVARH": "kVArh", "KW": "kW"}

REGISTER_CODE_LOOKUP = {"DAY": "00040", "NIGHT": "00206", "SINGLE": "00001"}

READ_TYPE_LOOKUP = {
    " ": "E",
    "E": "E",
    "N": "N",
    "S": "N",
}


def _handle_0461(headers, pre_record, record):
    parts = _chop_record(
        record,
        msn=11,
        unknown_1=2,
        prev_read_value=12,
        pres_read_value=12,
        coefficient=6,
        units=6,
        quantity=12,
        charge=6,
        prev_read_type=1,
        pres_read_type=1,
        mpan_core=13,
        mpan_top=8,
        register_code=19,
        pres_read_date=DATE_LENGTH,
        prev_read_date=DATE_LENGTH,
    )
    mpan_core = parse_mpan_core(parts["mpan_core"])
    headers["mpan_core"] = mpan_core
    units = CONSUMPTION_UNITS_LOOKUP[parts["units"].strip()]
    register_code = parts["register_code"].strip()
    if units == "kWh" and not (
        register_code.startswith("NOCHARGE") or register_code.endswith("XXXX")
    ):
        headers["kwh"] += Decimal(parts["quantity"])

    prev_read_date_str = parts["prev_read_date"].strip()
    if len(prev_read_date_str) > 0:
        tpr_code = REGISTER_CODE_LOOKUP[parts["register_code"].strip()]
        prev_type_code = READ_TYPE_LOOKUP[parts["prev_read_type"]]
        pres_type_code = READ_TYPE_LOOKUP[parts["pres_read_type"]]

        headers["reads"].append(
            {
                "msn": parts["msn"].strip(),
                "mpan": f"{parts['mpan_top']} {mpan_core}",
                "coefficient": Decimal(parts["coefficient"]),
                "units": units,
                "tpr_code": tpr_code,
                "prev_date": parse_date(parts["prev_read_date"]),
                "prev_value": Decimal(parts["prev_read_value"]),
                "prev_type_code": prev_type_code,
                "pres_date": parse_date(parts["pres_read_date"]),
                "pres_value": Decimal(parts["pres_read_value"]),
                "pres_type_code": pres_type_code,
            }
        )


def _handle_0470(headers, pre_record, record):
    pass


def _handle_0860(headers, pre_record, record):
    parts = _chop_record(
        record,
        metering_gbp=12,
        unknown_1=12,
        unknown_2=12,
        metering_date=DATE_LENGTH,
        description=80,
    )
    bd = headers["breakdown"]
    bd["metering-gbp"] += Decimal(parts["metering_gbp"]) / Decimal("100")


def _handle_1455(headers, pre_record, record):
    parts = _chop_record(
        record, ccl_kwh=13, unknown_1=10, ccl_rate=13, ccl_gbp=12, unkown_2=8
    )
    bd = headers["breakdown"]
    bd["ccl-kwh"] += Decimal(parts["ccl_kwh"])
    if "ccl-rate" in bd:
        ccl_rates = bd["ccl-rate"]
    else:
        ccl_rates = bd["ccl-rate"] = set()

    ccl_rates.add(Decimal(parts["ccl_rate"]) / Decimal("100"))
    bd["ccl-gbp"] += Decimal(parts["ccl_gbp"]) / Decimal("100")


def _handle_1460(headers, pre_record, record):
    parts = _chop_record(record, unknown_1=1, net=12, vat_rate=6, vat=12)
    net = Decimal(parts["net"]) / Decimal(100)
    vat_rate = int(Decimal(parts["vat_rate"]))
    vat = Decimal(parts["vat"]) / Decimal(100)

    vat_breakdown = headers["breakdown"]["vat"]
    try:
        vat_bd = vat_breakdown[vat_rate]
    except KeyError:
        vat_bd = vat_breakdown[vat_rate] = {"vat": Decimal("0"), "net": Decimal("0")}

    vat_bd["vat"] += vat
    vat_bd["net"] += net


def _handle_1500(headers, pre_record, record):
    parts = _chop_record(
        record,
        unknown_1=8,
        unknown_2=10,
        unknown_3=10,
        unknown_4=10,
        unknown_5=20,
        unknown_6=10,
        unknown_7=10,
        unknown_8=20,
        gross=12,
        net=12,
        vat=12,
    )
    breakdown = headers["breakdown"]
    net = Decimal("0.00") + Decimal(parts["net"]) / Decimal("100")
    gross = Decimal("0.00") + Decimal(parts["gross"]) / Decimal("100")
    if "late_payment" in headers:
        late_payment_gbp = headers["late_payment"]
        net += late_payment_gbp
        gross += late_payment_gbp
        breakdown["late-payment-gbp"] += late_payment_gbp
        del headers["late_payment"]

    return {
        "bill_type_code": "N",
        "mpan_core": headers["mpan_core"],
        "account": headers["account"],
        "reference": headers["reference"],
        "issue_date": headers["issue_date"],
        "start_date": headers["start_date"],
        "finish_date": headers["finish_date"],
        "kwh": headers["kwh"],
        "net": net,
        "vat": Decimal("0.00") + Decimal(parts["vat"]) / Decimal("100"),
        "gross": gross,
        "breakdown": breakdown,
        "reads": headers["reads"],
    }


def _handle_1600(headers, pre_record, record):
    parts = _chop_record(
        record,
        unknown_1=12,
        unknown_2=12,
        late_payment_fee=12,
        late_payment_date=DATE_LENGTH,
        description=50,
    )
    late_payment_gbp = Decimal(parts["late_payment_fee"]) / Decimal(100)

    headers["breakdown"]["late-payment-gbp"] += late_payment_gbp


def _handle_1700(headers, pre_record, record):
    pass


def _handle_2000(headers, pre_record, record):
    pass


def _handle_9999(headers, pre_record, record):
    pass


LINE_HANDLERS = {
    "0000": _handle_0000,
    "0050": _handle_0050,
    "0051": _handle_0051,
    "0100": _handle_0100,
    "0101": _handle_0101,
    "0460": _handle_0460,
    "0461": _handle_0461,
    "0470": _handle_0470,
    "0860": _handle_0860,
    "1455": _handle_1455,
    "1460": _handle_1460,
    "1500": _handle_1500,
    "1600": _handle_1600,
    "1700": _handle_1700,
    "2000": _handle_2000,
    "9999": _handle_9999,
}


class Parser:
    def __init__(self, f):
        self.f = StringIO(str(f.read(), "utf-8", errors="ignore"))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        headers = {}
        for self.line_number, line in enumerate(self.f, 1):
            pre_record, record_type, record = line[:80], line[80:84], line[84:]
            try:
                handler = LINE_HANDLERS[record_type]
            except KeyError:
                raise BadRequest(
                    f"Record type {record_type} not recognized on line "
                    f"{self.line_number} {line}"
                )

            try:
                bill = handler(headers, pre_record, record)
            except BadRequest as e:
                raise BadRequest(
                    f"Problem at line {self.line_number} {line}: {e.description}"
                )
            except BaseException as e:
                raise Exception(
                    f"Problem at line {self.line_number} {line}: {e}"
                ) from e
            if bill is not None:
                raw_bills.append(bill)

        return raw_bills



from decimal import Decimal

from sqlalchemy import select

from chellow.general_import import (
    _parse_breakdown,
    general_import_era,
    general_import_g_batch,
    general_import_g_bill,
    general_import_g_era,
    general_import_g_supply,
    general_import_llfc,
    general_import_site,
    general_import_supply,
)
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GBill,
    GContract,
    GDn,
    GReadingFrequency,
    GUnit,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Site,
    Source,
    Ssc,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_g_read_types,
    insert_g_reading_frequencies,
    insert_g_units,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, hh_format, to_utc, utc_datetime


def test_general_import_g_batch(mocker):
    sess = mocker.Mock()
    action = "insert"
    vals = ["CH4U", "batch 8883", "Apr 2019"]
    args = []
    general_import_g_batch(sess, action, vals, args)


def test_general_import_g_bill(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site_code = "22488"
    site = Site.insert(sess, site_code, "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "non core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract_name = "Fusion 2020"
    g_contract = GContract.insert_supplier(sess, g_contract_name, "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "87614362"
    msn = "hgeu8rhg"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    batch_name = "b1"
    g_contract.insert_g_batch(sess, batch_name, "batch 1")
    insert_bill_types(sess)
    sess.commit()

    action = "insert"
    vals = [
        g_contract_name,
        batch_name,
        mprn,
        "2019-09-08 00:00",
        "2019-10-01 00:00",
        "2019-10-31 23:30",
        "0.00",
        "0.00",
        "0.00",
        "77hwgtlll",
        "7876hrwlju",
        "N",
        "{}",
        "0",
    ]
    args = []
    general_import_g_bill(sess, action, vals, args)


def test_general_import_g_bill_reads(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site_code = "22488"
    site = Site.insert(sess, site_code, "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_code = "M3"
    g_unit = GUnit.get_by_code(sess, g_unit_code)
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "non core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract_name = "Fusion 2020"
    g_contract = GContract.insert_supplier(sess, g_contract_name, "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "87614362"
    msn = "hgeu8rhg"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    batch_reference = "b1"
    g_contract.insert_g_batch(sess, batch_reference, "batch 1")
    insert_bill_types(sess)
    insert_g_read_types(sess)
    sess.commit()
    msn = "88hgkdshjf"
    correction_factor_str = "1"
    correction_factor = Decimal("1")
    calorific_value_str = "39"
    calorific_value = Decimal(calorific_value_str)
    prev_value = Decimal("988")
    prev_date = utc_datetime(2019, 10, 1)
    prev_type_code = "E"
    pres_value = Decimal("1200")
    pres_date = utc_datetime(2019, 10, 31, 23, 30)
    pres_type_code = "A"

    action = "insert"
    vals = [
        g_contract_name,
        batch_reference,
        mprn,
        "2019-09-08 01:00",
        "2019-10-01 01:00",
        "2019-10-31 23:30",
        "0.00",
        "0.00",
        "0.00",
        "77hwgtlll",
        "7876hrwlju",
        "N",
        "{}",
        "0",
        msn,
        g_unit_code,
        correction_factor_str,
        calorific_value_str,
        hh_format(prev_date),
        str(prev_value),
        prev_type_code,
        hh_format(pres_date),
        str(pres_value),
        pres_type_code,
    ]
    args = []
    general_import_g_bill(sess, action, vals, args)

    g_bill = sess.scalars(select(GBill)).one()
    g_read = g_bill.g_reads[0]

    assert g_read.msn == msn
    assert g_read.g_unit == g_unit
    assert g_read.correction_factor == correction_factor
    assert g_read.calorific_value == calorific_value
    assert g_read.prev_value == prev_value
    assert g_read.prev_date == prev_date
    assert g_read.prev_type.code == prev_type_code
    assert g_read.pres_value == pres_value
    assert g_read.pres_date == pres_date
    assert g_read.pres_type.code == pres_type_code


def test_general_import_g_era_insert(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site_code = "22488"
    site = Site.insert(sess, site_code, "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "non core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract = GContract.insert_supplier(sess, "Fusion 2020", "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "87614362"
    msn = "hgeu8rhg"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    new_aq = "10"
    new_soq = "20"

    action = "insert"
    vals = [
        mprn,
        "2019-09-08 00:00",
        site_code,
        "{no change}",
        "1",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        new_aq,
        new_soq,
    ]
    args = []
    general_import_g_era(sess, action, vals, args)

    expected_args = [
        ("MPRN", "87614362"),
        ("Start Date", "2019-09-08 00:00"),
        ("Site Code", site_code),
        ("Meter Serial Number", "{no change}"),
        ("Correction Factor", "1"),
        ("Unit", "{no change}"),
        ("Supplier Contract Name", "{no change}"),
        ("Account", "{no change}"),
        ("Reading Frequency", "{no change}"),
        ("AQ", new_aq),
        ("SOQ", new_soq),
    ]

    assert args == expected_args


def test_general_import_g_era_insert_no_change(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "non core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract = GContract.insert_supplier(sess, "Fusion 2020", "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "87614362"
    msn = "hgeu8rhg"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )

    action = "insert"
    vals = [
        mprn,
        "2019-09-08 00:00",
        "{no change}",
        "{no change}",
        "1",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_g_era(sess, action, vals, args)
    expected_args = [
        ("MPRN", "87614362"),
        ("Start Date", "2019-09-08 00:00"),
        ("Site Code", "{no change}"),
        ("Meter Serial Number", "{no change}"),
        ("Correction Factor", "1"),
        ("Unit", "{no change}"),
        ("Supplier Contract Name", "{no change}"),
        ("Account", "{no change}"),
        ("Reading Frequency", "{no change}"),
        ("AQ", "{no change}"),
        ("SOQ", "{no change}"),
    ]

    assert args == expected_args


def test_general_import_g_era_update(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "non core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract = GContract.insert_supplier(sess, "Fusion 2020", "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "87614362"
    msn = "hgeu8rhg"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2010, 1, 1),
        None,
        msn,
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )

    action = "update"
    vals = [
        mprn,
        "2019-09-08 00:00",
        "{no change}",
        "{no change}",
        "A Mop Contract",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_g_era(sess, action, vals, args)
    expected_args = [
        ("MPRN", "87614362"),
        ("date", "2019-09-08 00:00"),
        ("Start Date", "{no change}"),
        ("Finish Date", "{no change}"),
        ("Meter Serial Number", "A Mop Contract"),
        ("Correction Factor", "{no change}"),
        ("Unit", "{no change}"),
        ("Supplier Contract Name", "{no change}"),
        ("Account", "{no change}"),
        ("Reading Frequency", "{no change}"),
        ("AQ", "{no change}"),
        ("SOQ", "{no change}"),
    ]

    assert args == expected_args


def test_general_import_g_supply_insert(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site_code = "22488"
    Site.insert(sess, site_code, "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone_code = "EA1"
    g_ldz.insert_g_exit_zone(sess, g_exit_zone_code)
    insert_g_units(sess)
    g_unit_code = "M3"
    GUnit.get_by_code(sess, g_unit_code)
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "non core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract_name = "Fusion 2020"
    GContract.insert_supplier(sess, g_contract_name, "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_code = "M"
    GReadingFrequency.get_by_code(sess, g_reading_frequency_code)
    mprn = "87614362"
    msn = "hgeu8rhg"
    aq = "10"
    soq = "20"
    g_supply_name = "1"
    start_date = "2019-09-08 00:00"
    finish_date = ""
    correction_factor = "1"
    account = "acc1"

    action = "insert"
    vals = [
        site_code,
        mprn,
        g_supply_name,
        g_exit_zone_code,
        start_date,
        finish_date,
        msn,
        correction_factor,
        g_unit_code,
        g_contract_name,
        account,
        g_reading_frequency_code,
        aq,
        soq,
    ]
    args = []
    general_import_g_supply(sess, action, vals, args)

    expected_args = [
        ("Site Code", site_code),
        ("MPRN", mprn),
        ("Supply Name", g_supply_name),
        ("Exit Zone", g_exit_zone_code),
        ("Start Date", start_date),
        ("Finish Date", finish_date),
        ("Meter Serial Number", msn),
        ("Correction Factor", correction_factor),
        ("Unit of Measurement", g_unit_code),
        ("Supplier Contract", g_contract_name),
        ("Account", account),
        ("Reading Frequency", g_reading_frequency_code),
        ("AQ", aq),
        ("SOQ", soq),
    ]

    assert args == expected_args


def test_parse_breakdown():
    breakdown_str = '{"date": 2009-05-12T03:00:00Z}'
    expected = {"date": utc_datetime(2009, 5, 12, 3)}
    actual = _parse_breakdown(breakdown_str)
    assert actual == expected


def test_general_import_era_insert(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "521", "Export (HV)", voltage_level, False, False, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )

    sess.commit()

    action = "insert"
    vals = [
        "22 7867 6232 781",
        "2020-10-01 00:00",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "Fusion Supplier 2000",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_era(sess, action, vals, args)


def test_general_import_era_insert_nhh(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "01", "nhh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "521", "Export (HV)", voltage_level, False, False, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    ssc_code = "0393"
    ssc = Ssc.insert(sess, ssc_code, "All", True, vf, None)

    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc_code,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )

    sess.commit()

    action = "insert"
    vals = [
        "22 7867 6232 781",
        "2020-10-01 00:00",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_era(sess, action, vals, args)


def test_general_import_era_update_hh(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "HH", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "521", "Export (HV)", voltage_level, False, False, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )
    action = "update"
    vals = [
        "2278676232781",
        "2019-09-08 00:00",
        "{no change}",
        "{no change}",
        "Fusion",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_era(sess, action, vals, args)


def test_general_import_era_update_nhh(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "01", "nhh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "521", "Export (HV)", voltage_level, False, False, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    ssc_code = "0393"
    ssc = Ssc.insert(sess, ssc_code, "All", True, vf, None)

    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc_code,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )
    action = "update"
    vals = [
        "2278676232781",
        "2019-09-08 00:00",
        "{no change}",
        "{no change}",
        "Fusion",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_era(sess, action, vals, args)


def test_general_import_llfc_insert(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "27"
    )
    insert_voltage_levels(sess)
    sess.commit()

    action = "insert"
    vals = [
        "27",
        "A1A",
        "LV:LV A Agg Band 0",
        "LV",
        "False",
        "True",
        "2020-10-21 00:00",
        "",
    ]
    args = []
    general_import_llfc(sess, action, vals, args)


def test_general_import_llfc_update(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "10"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "328",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2020, 4, 1)),
        None,
    )
    sess.commit()

    action = "update"
    vals = [
        "10",
        "328",
        "2020-04-01 00:00",
        "Reserved EHV 33kV - Import",
        "HV",
        "False",
        "True",
        "",
    ]
    args = []
    general_import_llfc(sess, action, vals, args)


def test_general_import_llfc_update_is_import_no_change(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "10"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "328",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2020, 4, 1)),
        None,
    )
    sess.commit()

    action = "update"
    vals = [
        "10",
        "328",
        "2020-04-01 00:00",
        "Reserved EHV 33kV - Import",
        "HV",
        "False",
        "{no change}",
        "",
    ]
    args = []
    general_import_llfc(sess, action, vals, args)


def test_general_import_llfc_update_valid_to_no_change(sess):
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "10"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "328",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(2020, 4, 1)),
        None,
    )
    sess.commit()

    action = "update"
    vals = [
        "10",
        "328",
        "2020-04-01 00:00",
        "Reserved EHV 33kV - Import",
        "HV",
        "False",
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_llfc(sess, action, vals, args)


def test_general_import_site_update(sess):
    site_code = "CI017"
    site_name = "Water Works"
    Site.insert(sess, site_code, site_name)
    sess.commit()

    action = "update"
    vals = [
        site_code,
        "{no change}",
        "{no change}",
    ]
    args = []
    general_import_site(sess, action, vals, args)

    site = sess.scalars(select(Site)).one()
    assert site.code == site_code
    assert site.name == site_name


def test_general_import_supply_insert_HH(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site_code = "CI017"
    Site.insert(sess, site_code, "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract_name = "Fusion"
    Contract.insert_mop(sess, mop_contract_name, participant, "", {}, vf, None, {})
    dc_contract_name = "Fusion DC 2000"
    Contract.insert_dc(sess, dc_contract_name, participant, "", {}, vf, None, {})
    Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop_code = "5"
    Cop.get_by_code(sess, cop_code)
    insert_comms(sess)
    comm_code = "GSM"
    Comm.get_by_code(sess, comm_code)
    supplier_contract_name = "Fusion 2000"
    Contract.insert_supplier(
        sess, supplier_contract_name, participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc_code = "845"
    mtc = Mtc.insert(sess, mtc_code, False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_code = "521"
    llfc = dno.insert_llfc(sess, llfc_code, "Imp", voltage_level, False, True, vf, None)
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source_code = "grid"
    Source.get_by_code(sess, source_code)
    insert_energisation_statuses(sess)
    energisation_status_code = "E"
    EnergisationStatus.get_by_code(sess, energisation_status_code)
    gsp_group_code = "_L"
    GspGroup.insert(sess, gsp_group_code, "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type_code = "H"
    supply_name = "Bob"
    start_date = vf
    msn = "khgsa;kh"

    sess.commit()

    action = "insert"
    vals = [
        site_code,
        source_code,
        "",
        supply_name,
        gsp_group_code,
        hh_format(start_date),
        "",
        mop_contract_name,
        "22 7867 6232 781",
        dc_contract_name,
        "22 7867 6232 781",
        msn,
        "22",
        "00",
        mtc_code,
        cop_code,
        comm_code,
        "",
        energisation_status_code,
        dtc_meter_type_code,
        "22 7867 6232 781",
        llfc_code,
        "0",
        supplier_contract_name,
        "22 7867 6232 781",
    ]
    args = []
    general_import_supply(sess, action, vals, args)


def test_general_import_supply_insert_NHH(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site_code = "CI017"
    Site.insert(sess, site_code, "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract_name = "Fusion"
    Contract.insert_mop(sess, mop_contract_name, participant, "", {}, vf, None, {})
    dc_contract_name = "Fusion DC 2000"
    Contract.insert_dc(sess, dc_contract_name, participant, "", {}, vf, None, {})
    pc_code = "01"
    pc = Pc.insert(sess, pc_code, "nhh", vf, None)
    insert_cops(sess)
    cop_code = "5"
    Cop.get_by_code(sess, cop_code)
    insert_comms(sess)
    comm_code = "GSM"
    Comm.get_by_code(sess, comm_code)
    supplier_contract_name = "Fusion 2000"
    Contract.insert_supplier(
        sess, supplier_contract_name, participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc_code = "845"
    mtc = Mtc.insert(sess, mtc_code, False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_code = "521"
    llfc = dno.insert_llfc(sess, llfc_code, "Imp", voltage_level, False, True, vf, None)
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source_code = "grid"
    Source.get_by_code(sess, source_code)
    insert_energisation_statuses(sess)
    energisation_status_code = "E"
    EnergisationStatus.get_by_code(sess, energisation_status_code)
    gsp_group_code = "_L"
    GspGroup.insert(sess, gsp_group_code, "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type_code = "N"
    supply_name = "Bob"
    start_date = vf
    msn = "khgsa;kh"
    ssc_code = "0393"
    ssc = Ssc.insert(sess, ssc_code, "All", True, vf, None)

    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    sess.commit()

    action = "insert"
    vals = [
        site_code,
        source_code,
        "",
        supply_name,
        gsp_group_code,
        hh_format(start_date),
        "",
        mop_contract_name,
        "22 7867 6232 781",
        dc_contract_name,
        "22 7867 6232 781",
        msn,
        "22",
        pc_code,
        mtc_code,
        cop_code,
        comm_code,
        ssc_code,
        energisation_status_code,
        dtc_meter_type_code,
        "22 7867 6232 781",
        llfc_code,
        "0",
        supplier_contract_name,
        "22 7867 6232 781",
    ]
    args = []
    general_import_supply(sess, action, vals, args)



import atexit
import collections
import threading
import traceback

from dateutil.relativedelta import relativedelta

import requests

from zish import loads

from chellow.models import Contract, RateScript, Session
from chellow.utils import (
    HH,
    hh_format,
    utc_datetime,
    utc_datetime_now,
    utc_datetime_parse,
)


bh_importer = None


def _run(log_f, sess):
    log_f("Starting to check bank holidays")
    contract = Contract.get_non_core_by_name(sess, "bank_holidays")
    contract_props = contract.make_properties()

    if contract_props.get("enabled", False):
        url_str = contract_props["url"]

        log_f(f"Downloading from {url_str}.")
        res = requests.get(url_str)
        log_f(" ".join(("Received", str(res.status_code), res.reason)))
        PREFIX = "DTSTART;VALUE=DATE:"
        hols = collections.defaultdict(list)
        for line in res.text.splitlines():
            if line.startswith(PREFIX):
                dt = utc_datetime_parse(line[-8:], "%Y%m%d")
                hols[dt.year].append(dt)

        for year in sorted(hols.keys()):
            year_start = utc_datetime(year, 1, 1)
            year_finish = year_start + relativedelta(years=1) - HH
            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date == year_start,
                )
                .first()
            )
            if rs is None:
                log_f(f"Adding a new rate script starting at {hh_format(year_start)}.")

                latest_rs = (
                    sess.query(RateScript)
                    .filter(RateScript.contract == contract)
                    .order_by(RateScript.start_date.desc())
                    .first()
                )

                contract.update_rate_script(
                    sess,
                    latest_rs,
                    latest_rs.start_date,
                    year_finish,
                    loads(latest_rs.script),
                )
                rs = contract.insert_rate_script(sess, year_start, {})

            script = {"bank_holidays": [v.strftime("%Y-%m-%d") for v in hols[year]]}

            contract.update_rate_script(sess, rs, rs.start_date, rs.finish_date, script)
            sess.commit()
            log_f(f"Updated rate script starting at {hh_format(year_start)}.")
    else:
        log_f(
            "The automatic importer is disabled. To enable it, edit the contract "
            "properties to set 'enabled' to True."
        )


class BankHolidayImporter(threading.Thread):
    def __init__(self):
        super(BankHolidayImporter, self).__init__(name="Bank Holiday Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = "proxy.host"
        self.PROXY_PORT_KEY = "proxy.port"
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                try:
                    with Session() as sess:
                        self.global_alert = None
                        _run(self.log, sess)
                except BaseException:
                    self.log(f"Outer problem {traceback.format_exc()}")
                    self.global_alert = (
                        "There's a problem with the Bank Holiday importer"
                    )
                finally:
                    self.lock.release()
                    self.log("Finished checking bank holidays.")

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def get_importer():
    return bh_importer


def startup():
    global bh_importer
    bh_importer = BankHolidayImporter()
    bh_importer.start()


@atexit.register
def shutdown():
    if bh_importer is not None:
        bh_importer.stop()



import csv
from io import StringIO

from utils import match_tables

import chellow.e.computer
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_87 import content
from chellow.utils import ct_datetime, to_utc


class Sess:
    def __init__(self, *results):
        self.it = iter(results)

    def query(self, *arg):
        return self

    def join(self, *arg):
        return self

    def order_by(self, *arg):
        return self

    def filter(self, *arg):
        return self

    def scalar(self, *arg):
        return next(self.it)

    def first(self, *arg):
        return next(self.it)

    def distinct(self, *arg):
        return next(self.it)


def test_summertime(mocker):
    start_date_ct = ct_datetime(2010, 5)
    finish_date_ct = ct_datetime(2010, 5, 31, 23, 30)
    start_date = to_utc(start_date_ct)
    finish_date = to_utc(finish_date_ct)
    contract_id = 1
    user = mocker.Mock()
    user.email_address = "sfreud"

    mock_c_months_u = mocker.patch(
        "chellow.reports.report_87.c_months_u", autospec=True
    )

    MockEra = mocker.patch("chellow.reports.report_87.Era", autospec=True)
    MockEra.finish_date = finish_date
    MockEra.start_date = start_date

    mocker.patch(
        "chellow.reports.report_87.contract_func",
        autospec=True,
        return_value=lambda: ["standing_gbp"],
    )

    f = StringIO()
    sess = Sess([], [])

    chellow.reports.report_87.create_csv(f, sess, start_date, finish_date, contract_id)

    mock_c_months_u.assert_called_with(
        start_year=start_date_ct.year,
        start_month=start_date_ct.month,
        finish_year=finish_date_ct.year,
        finish_month=finish_date_ct.month,
    )


def test_content_no_supplies(mocker, sess):
    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_87.open_file", return_value=mock_file)
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    vf = to_utc(ct_datetime(1996, 1, 1))
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)

    contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    sess.commit()

    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))

    content(start_date, finish_date, contract.id, user_id)


def test_content_one_supply(mocker, sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)

    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    supplier_charge_script = """
from chellow.utils import reduce_bill_hhs

def virtual_bill_titles():
    return [
        'net-gbp', 'vat-gbp', 'gross-gbp', 'sum-msp-kwh', 'sum-msp-gbp', 'problem'
    ]

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * 0.1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        vf,
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        vf,
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_87.open_file", return_value=mock_file)

    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))

    content(start_date, finish_date, imp_supplier_contract.id, user_id)

    mock_file.seek(0)
    sheet = csv.reader(mock_file)
    actual_table = list(sheet)

    expected_table = [
        [
            "mpan_core",
            "site_code",
            "site_name",
            "account",
            "from",
            "to",
            "energisation_status",
            "gsp_group",
            "dno",
            "era_start",
            "pc",
            "meter_type",
            "imp_is_substation",
            "imp_llfc_code",
            "imp_llfc_description",
            "exp_is_substation",
            "exp_llfc_code",
            "exp_llfc_description",
            "net-gbp",
            "vat-gbp",
            "gross-gbp",
            "sum-msp-kwh",
            "sum-msp-gbp",
            "problem",
        ],
        [
            "22 7867 6232 781",
            "CI017",
            "Water Works",
            "7748",
            "2020-01-01 00:00",
            "2020-01-31 23:30",
            "E",
            "_L",
            "22",
            "1996-01-01 00:00",
            "00",
            "C5",
            "False",
            "510",
            "PC 5-8 & HH HV",
            "",
            "",
            "",
            "0.0",
            "0",
            "0.0",
            "0",
            "0.0",
            "",
        ],
    ]
    match_tables(expected_table, actual_table)
